
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * General page setup
 */
#dt_example {
	font: 80%/1.45em "Lucida Grande", Verdana, Arial, Helvetica, sans-serif;
	margin: 0;
	padding: 0;
	color: #333;
	background-color: #fff;
}


#dt_example #container {
	width: 800px;
	margin: 30px auto;
	padding: 0;
}


#dt_example #footer {
	margin: 50px auto 0 auto;
	padding: 0;
}

#dt_example #demo {
	margin: 30px auto 0 auto;
}

#dt_example .demo_jui {
	margin: 30px auto 0 auto;
}

#dt_example .big {
	font-size: 1.3em;
	font-weight: bold;
	line-height: 1.6em;
	color: #4E6CA3;
}

#dt_example .spacer {
	height: 20px;
	clear: both;
}

#dt_example .clear {
	clear: both;
}

#dt_example pre {
	padding: 15px;
	background-color: #F5F5F5;
	border: 1px solid #CCCCCC;
}

#dt_example h1 {
	margin-top: 2em;
	font-size: 1.3em;
	font-weight: normal;
	line-height: 1.6em;
	color: #4E6CA3;
	border-bottom: 1px solid #B0BED9;
	clear: both;
}

#dt_example h2 {
	font-size: 1.2em;
	font-weight: normal;
	line-height: 1.6em;
	color: #4E6CA3;
	clear: both;
}

#dt_example a {
	color: #0063DC;
	text-decoration: none;
}

#dt_example a:hover {
	text-decoration: underline;
}

#dt_example ul {
	color: #4E6CA3;
}

.css_right {
	float: right;
}

.css_left {
	float: left;
}

.demo_links {
	float: left;
	width: 50%;
	margin-bottom: 1em;
}