/*
 * Namespace DTCR - "DataTables ColReorder" plug-in
 */

table.DTCR_clonedTable {
	background-color: white;
	z-index: 202;
}

div.DTCR_pointer {
	width: 1px;
	background-color: #0259C4;
	z-index: 201;
}

body.alt div.DTCR_pointer {
	margin-top: -15px;
	margin-left: -9px;
	width: 18px;
	background: url('../images/insert.png') no-repeat top left;
}